# Error API path
ERROR_API_PATH = '/v1/errors'

# Transactions API path
TRANSACTIONS_API_PATH = '/v1/transactions'

TIMESTAMP_FORMAT = '%Y-%m-%dT%H:%M:%S.%fZ'

KEYWORD_MAX_LENGTH = 1024

HTTP_WITH_BODY = {'POST', 'PUT', 'PATCH', 'DELETE'}
